@@ -220,17 +220,15 @@ certificate_button_clicked_cb (GtkButton *button,
                                gpointer   user_data)
 {
   EphySecurityPopover *popover = EPHY_SECURITY_POPOVER (user_data);
-  GtkWidget *dialog;
+  AdwDialog *dialog;
 
-  dialog = ephy_certificate_dialog_new (GTK_WINDOW (gtk_widget_get_root (GTK_WIDGET (popover))),
-                                        popover->address,
+  dialog = ephy_certificate_dialog_new (popover->address,
                                         popover->certificate,
                                         popover->tls_errors,
                                         popover->security_level);
-  gtk_window_set_destroy_with_parent (GTK_WINDOW (dialog), TRUE);
 
+  adw_dialog_present (dialog, GTK_WIDGET (gtk_widget_get_root (GTK_WIDGET (popover))));
   gtk_popover_popdown (GTK_POPOVER (popover));
-  gtk_window_present (GTK_WINDOW (dialog));
 }
 
 static void
@@ -362,7 +360,7 @@ ephy_security_popover_class_init (EphySecurityPopoverClass *klass)
 }
 
 G_GNUC_BEGIN_IGNORE_DEPRECATIONS
-static gboolean
+static void
 on_ad_combobox_changed (GtkComboBox         *widget,
                         EphySecurityPopover *popover)
 {
@@ -375,7 +373,7 @@ on_ad_combobox_changed (GtkComboBox         *widget,
 
   origin = ephy_uri_to_security_origin (popover->address);
   if (!origin)
-    return FALSE;
+    return;
 
   permissions_manager = ephy_embed_shell_get_permissions_manager (ephy_embed_shell_get_default ());
 
@@ -387,7 +385,7 @@ on_ad_combobox_changed (GtkComboBox         *widget,
                                            origin,
                                            permission);
 
-  return FALSE;
+  gtk_popover_popdown (GTK_POPOVER (popover));
 }
 G_GNUC_END_IGNORE_DEPRECATIONS
 
@@ -423,6 +421,8 @@ handle_permission_combobox_changed (EphySecurityPopover *popover,
                                            permission_type,
                                            origin,
                                            permission);
+
+  gtk_popover_popdown (GTK_POPOVER (popover));
 }
 
 G_GNUC_BEGIN_IGNORE_DEPRECATIONS
